#!/usr/bin/env python3

# Copyright 2019 Open Source Robotics Foundation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from collections import OrderedDict
import os
import sys

import pkg_resources
import pkgutil
import subprocess
import tempfile

import em

import docker
import requests
import pexpect


class DockerImageGenerator(object):
    def __init__(self, active_extensions, cliargs, base_image):
        self.built = False
        self.cliargs = cliargs
        self.active_extensions = active_extensions

        self.dockerfile = generate_dockerfile(active_extensions, self.cliargs, base_image)
        # print(df)
        self.image_name = "rocker_" + base_image
        if self.active_extensions:
            self.image_name += "_%s" % '_'.join([e.name for e in active_extensions])

    
    def build(self, **kwargs):
        with tempfile.TemporaryDirectory() as td:
            df = os.path.join(td, 'Dockerfile')
            print("Writing dockerfile to %s" % df)
            with open(df, 'w') as fh:
                fh.write(self.dockerfile)
            print('vvvvvv')
            print(self.dockerfile)
            print('^^^^^^')
            arguments = {}
            arguments['path'] = td
            arguments['rm'] = True
            arguments['decode'] = True
            arguments['nocache'] = kwargs.get('nocache', False)
            arguments['tag'] = self.image_name
            print("Building docker file with arguments: ", arguments)
            try:
                docker_client = docker.APIClient()
                success_detected = False
                for line in docker_client.build(**arguments):
                    output = line.get('stream', '').rstrip()
                    if not output:
                        # print("non stream data", line)
                        continue
                    print("building > %s" % output)
                    if output.startswith("Successfully tagged") and self.image_name in output:
                        success_detected = True
                if success_detected:
                        self.built = True
                        return True
 
            except docker.errors.APIError as ex:
                print("Docker build failed\n", ex)
                print(ex.output)
                return False
        # Unknown error 
        return False

    def run(self, command=[], **kwargs):
        if not self.built:
            print("Cannot run if build has not passed.")
            return False

        for e in self.active_extensions:
            try:
                e.precondition_environment(self.cliargs)
            except subprocess.CalledProcessError as ex:
                print("Failed to precondition for extension [%s] with error: %s\ndeactivating" % (e.get_name(), ex))
                # TODO(tfoote) remove the extension from the list


        docker_args = ''

        network = kwargs.get('network', False)
        if network:
            docker_args += ' --network %s ' % network

        devices = kwargs.get('devices', None)
        if devices:
            for device in devices:
                if not os.path.exists(device):
                    print("ERROR device %s doesn't exist. Skipping" % device)
                    continue
                docker_args += ' --device %s ' % device

        for e in self.active_extensions:
            docker_args += e.get_docker_args(self.cliargs)

        image = self.image_name
        cmd="docker run -it \
  --rm \
  %(docker_args)s \
  %(image)s %(command)s" % locals()
#   $DOCKER_OPTS \
        try:
            if kwargs.get('execute', False):
                print("Executing command: ")
                print(cmd)
                p = pexpect.spawn(cmd)
                p.interact()
            else:
                print("Run this command: \n\n\n")
                print(cmd)
        except subprocess.CalledProcessError as ex:
            print("Docker run failed\n", ex)
            print(ex.output)
            return False


def generate_dockerfile(extensions, args_dict, base_image):
    dockerfile_str = ''
    for el in extensions:
        dockerfile_str += '# Preamble from extension [%s]\n' % el.name
        dockerfile_str += el.get_preamble(args_dict) + '\n'
    dockerfile_str += '\nFROM %s\n' % base_image
    for el in extensions:
        dockerfile_str += '# Snippet from extension [%s]\n' % el.name
        dockerfile_str += el.get_snippet(args_dict) + '\n'
    return dockerfile_str


def main():

    parser = argparse.ArgumentParser(description='A tool for running docker with extra options')
    parser.add_argument('image', nargs='+')
    parser.add_argument('--execute', action='store_true')
    parser.add_argument('--nocache', action='store_true')
    parser.add_argument('--pull', action='store_true')
    parser.add_argument('--network', choices=['bridge', 'host', 'overlay', 'none'])
    parser.add_argument('--devices', nargs='*')

    unordered_plugins = {
    entry_point.name: entry_point.load()
    for entry_point
    in pkg_resources.iter_entry_points('rocker.extensions')
    }
    # Order plugins by extension point name for consistent ordering below
    plugin_names = list(unordered_plugins.keys())
    plugin_names.sort()
    plugins = OrderedDict([(k, unordered_plugins[k]) for k in plugin_names])

    print("Plugins found: %s" % [p.get_name() for p in plugins.values()])
    for p in plugins.values():
        p.register_arguments(parser)

    args = parser.parse_args()
    args_dict = vars(args)
    
    active_extensions = [e() for e in plugins.values() if args_dict.get(e.get_name())]
    # Force user to end if present otherwise it will 
    active_extensions.sort(key=lambda e:e.get_name().startswith('user'))
    print("Active extensions %s" % [e.get_name() for e in active_extensions])

    base_image = args.image[0]

    if args.pull:
        docker_client = docker.APIClient()
        try:
            print("Pulling image %s" % base_image)
            for line in docker_client.pull(base_image, stream=True):
                print(line)
        except docker.errors.APIError as ex:
            print('Pull of %s failed: %s' % (base_image, ex))
            pass
    dig = DockerImageGenerator(active_extensions, args_dict, base_image)
    dig.build(**vars(args))
    dig.run(command=' '.join(args.image[1:]), **args_dict)
    return 1


if __name__ == '__main__':
    sys.exit(main() or 0)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import docker
import unittest
import pexpect

from io import BytesIO as StringIO

class NvidiaTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        client = docker.from_env()
        dockerfile = """
FROM ubuntu:xenial

RUN apt-get update && apt-get install mesa-utils -y && apt-get clean

CMD glxgears
"""
        iof = StringIO(dockerfile.encode())
        im = client.images.build(fileobj = iof, tag='rocker_test_glxgears')
    
    def test_tripwire(self):
        self.assertTrue(True)


# Copyright 2019 Open Source Robotics Foundation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grp
import os
import em
import getpass
import pkgutil
from pathlib import Path

import subprocess

import distro

def name_to_argument(name):
    return '--%s' % name.replace('_', '-')

class RockerExtension(object):

    def precondition_environment(self, cliargs):
        """Modify the local environment such as setup tempfiles"""
        pass

    def validate_environment(self, cliargs):
        """ Check that the environment is something that can be used.
        This will check that we're on the right base OS and that the 
        necessary resources are available, like hardware."""
        pass

    def get_preamble(self, cliargs):
        return ''

    def get_snippet(self, cliargs):
        return ''

    def get_name(self, cliargs):
        raise NotImplementedError
    
    def get_docker_args(self, cliargs):
        return ''

    @staticmethod
    def register_arguments(parser):
        raise NotImplementedError

class DevHelpers(RockerExtension):
    @staticmethod
    def get_name():
        return 'dev_helpers'

    def __init__(self):
        self.env_subs = None
        self.name = DevHelpers.get_name()


    def get_environment_subs(self):
        if not self.env_subs:
            self.env_subs = {}
            self.env_subs['user_id'] = os.getuid()
            self.env_subs['username'] = getpass.getuser()
        return self.env_subs

    def get_preamble(self, cliargs):
        return ''

    def get_snippet(self, cliargs):
        snippet = pkgutil.get_data('rocker', 'templates/%s_snippet.Dockerfile.em' % self.name).decode('utf-8')
        return em.expand(snippet, self.get_environment_subs())

    @staticmethod
    def register_arguments(parser):
        parser.add_argument(name_to_argument(DevHelpers.get_name()),
            action='store_true',
            help="add development tools emacs and byobu to your environment")


class Nvidia(RockerExtension):
    @staticmethod
    def get_name():
        return 'nvidia'

    def __init__(self):
        self.env_subs = None
        self.name = Nvidia.get_name()
        self.xauth = '/tmp/.docker.xauth'
        self.supported_versions = ['16.04', '18.04']


    def get_environment_subs(self):
        if not self.env_subs:
            self.env_subs = {}
            self.env_subs['user_id'] = os.getuid()
            self.env_subs['username'] = getpass.getuser()
            self.env_subs['DISPLAY'] = os.getenv('DISPLAY')
            self.env_subs['distro_id'] = distro.id()
            if self.env_subs['distro_id'] != 'ubuntu':
                print("WARNING distro id %s not supported by Nvidia " % self.env_subs['distro_id'])
                sys.exit(1)
            self.env_subs['distro_version'] = distro.version()
            if self.env_subs['distro_version'] not in self.supported_versions:
                print("WARNING distro version %s not in supported list by Nvidia " % self.env_subs['distro_id'])
                sys.exit(1)
                # TODO(tfoote) add a standard mechanism for checking preconditions and disabling plugins

        return self.env_subs

    def get_preamble(self, cliargs):
        preamble = pkgutil.get_data('rocker', 'templates/%s_preamble.Dockerfile.em' % self.name).decode('utf-8')
        return em.expand(preamble, self.get_environment_subs())

    def get_snippet(self, cliargs):
        snippet = pkgutil.get_data('rocker', 'templates/%s_snippet.Dockerfile.em' % self.name).decode('utf-8')
        return em.expand(snippet, self.get_environment_subs())

    def get_docker_args(self, cliargs):
        xauth = self.xauth
        return "  -e DISPLAY -e TERM \
  -e QT_X11_NO_MITSHM=1 \
  -e XAUTHORITY=%(xauth)s -v %(xauth)s:%(xauth)s \
  -v /tmp/.X11-unix:/tmp/.X11-unix \
  -v /etc/localtime:/etc/localtime:ro \
  --runtime=nvidia \
  --security-opt seccomp=unconfined" % locals()

    def precondition_environment(self, cliargs):
        xauth = self.xauth
        display = os.getenv('DISPLAY')
        # Make sure processes in the container can connect to the x server
        # Necessary so gazebo can create a context for OpenGL rendering (even headless)
        if not os.path.exists(self.xauth): #if [ ! -f $XAUTH ]
            Path(self.xauth).touch()
            # print("touched %s" % xauth)
        cmd = 'xauth nlist %(display)s | sed -e \'s/^..../ffff/\' | xauth -f %(xauth)s nmerge -' % locals()
        # print("runnning %s" % cmd)
        try:
            subprocess.check_call(cmd, shell=True)
        except subprocess.CalledProcessError as ex:
            print("Failed setting up XAuthority with command %s" % cmd)
            raise ex
        # import stat
        # Path(xauth).chmod(stat.S_IROTH) 

    @staticmethod
    def register_arguments(parser):
        parser.add_argument(name_to_argument(Nvidia.get_name()),
            action='store_true',
            help="Enable nvidia")


class PulseAudio(RockerExtension):
    @staticmethod
    def get_name():
        return 'pulse'

    def __init__(self):
        self.env_subs = None
        self.name = PulseAudio.get_name()


    def get_environment_subs(self):
        if not self.env_subs:
            self.env_subs = {}
            self.env_subs['user_id'] = os.getuid()
            self.env_subs['XDG_RUNTIME_DIR'] = os.getenv('XDG_RUNTIME_DIR')
            self.env_subs['audio_group_id'] = grp.getgrnam('audio').gr_gid
        return self.env_subs

    def get_preamble(self, cliargs):
        return ''

    def get_snippet(self, cliargs):
        snippet = pkgutil.get_data('rocker', 'templates/%s_snippet.Dockerfile.em' % self.name).decode('utf-8')
        return em.expand(snippet, self.get_environment_subs())

    def get_docker_args(self, cliargs):
        args = ' -v /run/user/%(user_id)s/pulse:/run/user/%(user_id)s/pulse --device /dev/snd '\
        ' -e PULSE_SERVER=unix:%(XDG_RUNTIME_DIR)s/pulse/native -v %(XDG_RUNTIME_DIR)s/pulse/native:%(XDG_RUNTIME_DIR)s/pulse/native --group-add %(audio_group_id)s '
        return args % self.get_environment_subs()

    def precondition_environment(self, cliargs):
        pass

    @staticmethod
    def register_arguments(parser):
        parser.add_argument(name_to_argument(PulseAudio.get_name()),
            action='store_true',
            help="mount pulse audio devices")


class HomeDir(RockerExtension):
    @staticmethod
    def get_name():
        return 'home'

    def __init__(self):
        self.env_subs = None
        self.name = HomeDir.get_name()

    def get_docker_args(self, cliargs):
        return ' -v %s:%s ' % (Path.home(), Path.home())

    @staticmethod
    def register_arguments(parser):
        parser.add_argument(name_to_argument(HomeDir.get_name()),
            action='store_true',
            help="mount the users home directory")


class User(RockerExtension):
    @staticmethod
    def get_name():
        return 'user'

    def get_environment_subs(self):
        if not self.env_subs:
            self.env_subs = {}
            self.env_subs['user_id'] = os.getuid()
            self.env_subs['username'] = getpass.getuser()
        return self.env_subs

    def __init__(self):
        self.env_subs = None
        self.name = User.get_name()

    def get_snippet(self, cliargs):
        snippet = pkgutil.get_data('rocker', 'templates/%s_snippet.Dockerfile.em' % self.name).decode('utf-8')
        return em.expand(snippet, self.get_environment_subs())

    @staticmethod
    def register_arguments(parser):
        parser.add_argument(name_to_argument(User.get_name()),
            action='store_true',
            help="mount the users home directory")





#!/usr/bin/env python3

from setuptools import setup

install_requires = [
    'distro',
    'docker',
    'empy',
    'pexpect',
    'requests',
]

kwargs = {
    'name': 'rocker',
    'version': '0.0.1',
    'packages': ['rocker'],
    'package_dir': {'': 'src'},
    'package_data': {'rocker': ['templates/*.em']},
    'entry_points': {
        'console_scripts': [
            'rocker = rocker.cli:main',
	    ],
        'rocker.extensions': [
            'dev_helpers = rocker.extensions:DevHelpers',
            'nvidia = rocker.extensions:Nvidia',
            'pulse = rocker.extensions:PulseAudio',
            'home = rocker.extensions:HomeDir',
            'user = rocker.extensions:User',
        ]
	},
    'author': 'Tully Foote',
    'author_email': 'tfoote@osrfoundation.org',
    'keywords': ['Docker'],
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License'
    ],
    'description': 'A tool to run docker containers with customized extras',
    'long_description': 'A tool to run docker containers with customized extra added like nvidia gui support overlayed.',
    'license': 'Apache License 2.0',
    'python_requires': '>=3.0',

    'install_requires': install_requires,
    'url': 'https://github.com/osrf/rocker'
}

setup(**kwargs)

